from typing import Union

from pydantic import ValidationError
from returns.functions import raise_exception
from returns.result import Failure, ResultE

from kaiba.mapper import map_data
from kaiba.models.kaiba_object import KaibaObject


def process(
    input_data: dict,
    configuration: dict,
) -> ResultE[Union[list, dict]]:
    """Validate configuration then process data."""
    try:
        cfg = KaibaObject(**configuration)
    except ValidationError as ve:
        return Failure(ve)

    return map_data(input_data, cfg)


def process_raise(
    input_data: dict,
    configuration: dict,
) -> Union[list, dict]:
    """Call Process and unwrap value if no error, otherwise raise."""
    return process(
        input_data,
        configuration,
    ).alt(
        raise_exception,
    ).unwrap()



from typing import List

from pydantic import ConfigDict
from pydantic.types import StrictBool

from kaiba.models.attribute import Attribute
from kaiba.models.base import KaibaBaseModel
from kaiba.models.iterator import Iterator


class BranchingObject(KaibaBaseModel):
    """Lets you branch out attribute mappings.

    The branching object is a special kind of object where you can have
    different attributes per instance.

    ie:
    {
        "extra_fields": [
            {
                "name": "field1",
                "value": "value at path x",
            },
            {
                "name": "field2",
                "value": "value at path y"
            },
            {
                "other_key": "whatever you want",
            }
        ]
    }

    Whereas normal objects you'd have to have both `name`, `value` and
    `other_key` in all instances.
    """

    name: str
    array: StrictBool = False
    iterators: List[Iterator] = []
    branching_attributes: List[List[Attribute]] = []
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'name': 'extra_fields',
                'array': True,
                'branching_attributes': [
                    [
                        {
                            'name': 'field_name',
                            'default': 'amount',
                        },
                        {
                            'name': 'field_data',
                            'data_fetchers': [
                                {
                                    'path': ['path', 'to', 'amount'],
                                },
                            ],
                        },
                    ],
                    [
                        {
                            'name': 'field_name',
                            'default': 'currency',
                        },
                        {
                            'name': 'field_data',
                            'data_fetchers': [
                                {
                                    'path': ['path', 'to', 'currency'],
                                },
                            ],
                            'default': 'NOK',
                        },
                    ],
                ],
            },
        ],
    })



from typing import List, Pattern, Union

from pydantic import ConfigDict
from pydantic.types import StrictInt

from kaiba.models.base import KaibaBaseModel


class Regex(KaibaBaseModel):
    """Use regular expression on data found by data_fetchers."""

    expression: Pattern
    group: Union[StrictInt, List[StrictInt]] = 0
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'expression': '[a-z]+',
            },
            {
                'expression': '([a-z])',
                'group': [0, 3, 4],
            },
        ],
    })



from enum import Enum
from typing import Optional

from pydantic import ConfigDict, Field

from kaiba.models.base import AnyType, KaibaBaseModel


class Conditions(str, Enum):  # noqa: WPS600
    """Conditions for if statements."""

    IS = 'is'  # noqa: WPS115
    NOT = 'not'  # noqa: WPS115
    IN = 'in'  # noqa: WPS115
    CONTAINS = 'contains'  # noqa: WPS115


class IfStatement(KaibaBaseModel):
    """If statements lets you conditionally change data."""

    condition: Conditions
    target: Optional[AnyType] = Field(...)  # ... = required but allow None
    then: Optional[AnyType] = Field(...)  # ... = required but allow Nones
    otherwise: Optional[AnyType] = None  # Should be any valid json value
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'condition': 'is',
                'target': 'target value',
                'then': 'was target value',
            },
            {
                'condition': 'not',
                'target': 'target_value',
                'then': 'was not target value',
                'otherwise': 'was target value',
            },
            {
                'condition': 'in',
                'target': ['one', 'of', 'these'],
                'then': 'was either one, of or these',
                'otherwise': 'was none of those',
            },
            {
                'condition': 'in',
                'target': 'a substring of this will be true',
                'then': 'substrings also work',
                'otherwise': 'was not a substring of target',
            },
            {
                'condition': 'contains',
                'target': 'value',
                'then': 'value was a substring of input',
                'otherwise': 'value was not in the input',
            },
        ],
    })



from typing import Optional

from pydantic import ConfigDict, Field
from pydantic.types import StrictInt

from kaiba.models.base import KaibaBaseModel


class Slicing(KaibaBaseModel):
    """Slice from inclusive to exclusive like python slice."""

    slice_from: StrictInt = Field(alias='from')
    slice_to: Optional[StrictInt] = Field(None, alias='to')
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'from': 3,
            },
            {
                'from': -5,
            },
            {
                'from': 3,
                'to': 10,
            },
        ],
    })



from decimal import Decimal
from typing import Union

from pydantic import BaseModel, ConfigDict
from pydantic.types import StrictBool, StrictInt, StrictStr

AnyType = Union[StrictStr, StrictInt, StrictBool, Decimal, list, dict]
StrInt = Union[StrictStr, StrictInt]


class KaibaBaseModel(BaseModel):
    """Base model that forbids non defined attributes."""

    model_config = ConfigDict(extra='forbid')



from __future__ import annotations

import re
from decimal import Decimal
from typing import Any, List, Optional, Union

from returns.pipeline import flow
from returns.pointfree import bind
from returns.result import Failure, ResultE, safe

from kaiba.casting import get_casting_function
from kaiba.models.base import AnyType
from kaiba.models.casting import Casting
from kaiba.models.if_statement import Conditions, IfStatement
from kaiba.models.regex import Regex
from kaiba.models.slicing import Slicing

ValueTypes = (str, int, float, bool, Decimal)


@safe
def apply_if_statements(
    if_value: Optional[AnyType],
    statements: List[IfStatement],
) -> Optional[AnyType]:
    """Apply if statements to a value.

    :param if_value: The value to use when evaluating if statements
    :type if_value: AnyType

    :param statements: :term:`statements` collection of if operations
    :type statements: List[Dict[str, Any]]

    :return: Success/Failure containers
    :rtype: AnyType

    one if object looks like this

    .. code-block:: json

        {
            "condition": "is",
            "target": "foo",
            "then": "bar",
            "otherwise": "no foo" - optional
        }

    The if statements are chained so that the next works on the output of the
    previous. If no "otherwise" is provided then the original value or value
    from the previous operation will be returned.

    Example
        >>> apply_if_statements(
        ...     '1', [
        ...         IfStatement(
        ...             **{'condition': 'is', 'target': '1', 'then': '2'}
        ...         )
        ...     ],
        ... ).unwrap() == '2'
        True
        >>> apply_if_statements(
        ...     'a',
        ...     [IfStatement(**{
        ...         'condition': 'is',
        ...         'target': '1',
        ...         'then': '2',
        ...         'otherwise': '3'
        ...     })],
        ... ).unwrap() == '3'
        True

    """
    for statement in statements:

        if_value = _apply_statement(
            if_value, statement,
        )

    if if_value is None:
        raise ValueError('If statement failed or produced `None`')

    return if_value


def _apply_statement(
    if_value: Optional[AnyType],
    statement: IfStatement,
) -> Optional[AnyType]:
    evaluation: bool = False

    condition = statement.condition
    target = statement.target

    if condition == Conditions.IS:
        evaluation = if_value == target

    if condition == Conditions.NOT:
        evaluation = if_value != target

    if condition == Conditions.IN:
        evaluation = if_value in target  # type: ignore

    if condition == Conditions.CONTAINS:
        list_or_dict = isinstance(if_value, (dict, list))
        evaluation = list_or_dict and target in if_value  # type: ignore
        evaluation = evaluation or not list_or_dict and str(target) in str(if_value)  # noqa: E501 E262
    if evaluation:
        return statement.then

    return statement.otherwise or if_value


@safe
def apply_separator(
    mapped_values: List[AnyType],
    separator: str,
) -> AnyType:
    """Apply separator between the values of a List[Any].

    :param mapped_values: The list of values to join with the separator
    :type mapped_values: List[AnyType]

    :param separator: :term:`separator` value to join mapped_values list with
    :type separator: str

    :return: Success/Failure containers
    :rtype: AnyType

    Example
        >>> from returns.pipeline import is_successful
        >>> apply_separator(['a', 'b', 'c'], ' ').unwrap()
        'a b c'
        >>> apply_separator([1, 'b', True], ' ').unwrap()
        '1 b True'
        >>> is_successful(apply_separator([], ' '))
        False

    """
    if not mapped_values:
        raise ValueError('mapped_values is empty')

    if len(mapped_values) == 1:
        return mapped_values[0]

    return separator.join([str(mapped) for mapped in mapped_values])


def apply_slicing(
    value_to_slice: Optional[Any],
    slicing: Slicing | None,
) -> Optional[AnyType]:
    """Slice value from index to index.

    :param value_to_slice: The value to slice
    :type value_to_slice: Any

    :param slicing: :term:`slicing` object
    :type slicing: dict


    :return: Success/Failure containers
    :rtype: Any

    Example
        >>> apply_slicing('123', Slicing(**{'from': 1}))
        '23'
        >>> apply_slicing('test', Slicing(**{'from': 1, 'to': 3}))
        'es'
    """
    if value_to_slice is None:
        return value_to_slice

    if not slicing:
        return value_to_slice

    if not isinstance(value_to_slice, list):
        value_to_slice = str(value_to_slice)

    return value_to_slice[slicing.slice_from:slicing.slice_to]


@safe
def apply_regex(  # noqa: WPS212, WPS234
    value_to_match: Optional[AnyType],
    regex: Regex | None,
) -> Union[List[AnyType], AnyType, None]:
    r"""Match value by a certain regex pattern.

    :param value_to_match: The value to match
    :type value_to_match: AnyType

    :param regex: :term: `matching` object which has parameters for
        regex match
    :type regex: Regex

    :return: Success/Failure container
    :rtype: AnyType

    Example
        >>> apply_regex(
        ...     'abcdef',
        ...     Regex(**{'expression': '(?<=abc)def'})
        ... ).unwrap()
        'def'
        >>> apply_regex(
        ...     'Isaac Newton, physicist',
        ...     Regex(**{'expression': r'(\w+)', 'group': 1}),
        ... ).unwrap()
        'Newton'
        >>> apply_regex(
        ...     'Isaac Newton, physicist',
        ...     Regex(**{'expression': r'(\w+)', 'group': [1, 2]}),
        ... ).unwrap()
        ['Newton', 'physicist']
        >>> apply_regex(None, Regex(**{'expression': 'a+'})).unwrap()
        >>> apply_regex('Open-source matters', None).unwrap()
        'Open-source matters'
    """
    if value_to_match is None:
        return value_to_match

    if not regex or not regex.expression:
        return value_to_match

    pattern = regex.expression
    groups = re.finditer(
        pattern,
        value_to_match,  # type: ignore
    )
    matches: list = [gr.group(0) for gr in groups]
    num_group: Union[int, List[int]] = regex.group
    if isinstance(num_group, list):
        if not num_group:
            return matches
        return [matches[ind] for ind in num_group]  # typing: ignore
    return matches[num_group]


def apply_casting(
    value_to_cast: Optional[AnyType],
    casting: Casting,
) -> ResultE[AnyType]:
    """Casting one type of code to another.

    :param casting: :term:`casting` object
    :type casting: dict

    :param value_to_cast: The value to cast to casting['to']
    :type value_to_cast: AnyType

    :return: Success/Failure containers
    :rtype: AnyType

    Example
        >>> apply_casting('123', Casting(**{'to': 'integer'})).unwrap()
        123
        >>> apply_casting('123.12', Casting(**{'to': 'decimal'})).unwrap()
        Decimal('123.12')
    """
    if value_to_cast is None:
        return Failure(ValueError('value_to_cast is empty'))

    return flow(
        casting.to,
        get_casting_function,
        bind(  # type: ignore
            lambda function: function(  # type: ignore
                value_to_cast, casting.original_format,
            ),
        ),
    )


@safe
def apply_default(
    mapped_value: Optional[AnyType] = None,
    default: Optional[AnyType] = None,
) -> AnyType:
    """Apply default value if exists and if mapped value is None.

    :param mapped_value: If this value is *None* the default value is returned
    :type mapped_value: Optional[AnyType]

    :param default: :term:`default` value to return if mapped value is None
    :type default: Optional[AnyType]

    :return: Success/Failure containers
    :rtype: AnyType

    If default *is not* none and mapped_value *is* None then return default
    else if mapped_value is not in accepted ValueTypes then throw an error
    else return mapped value

    Example
        >>> apply_default('test', None).unwrap() == 'test'
        True
        >>> apply_default('nope', 'test').unwrap() == 'nope'
        True
    """
    if default is not None:
        if mapped_value is None:
            return default

    if mapped_value is None:
        raise ValueError('Default value should not be `None`')

    if not isinstance(mapped_value, ValueTypes):
        raise ValueError('Unable to give default value')

    return mapped_value



from typing import Any, Dict, List, Union

from returns.curry import partial
from returns.pipeline import flow, is_successful
from returns.pointfree import bind, lash, map_
from returns.result import ResultE, Success, safe

from kaiba.collection_handlers import fetch_data_by_keys
from kaiba.functions import (
    apply_casting,
    apply_default,
    apply_if_statements,
    apply_regex,
    apply_separator,
    apply_slicing,
)
from kaiba.models.attribute import Attribute
from kaiba.models.base import AnyType
from kaiba.models.data_fetcher import DataFetcher


def handle_data_fetcher(
    collection: Union[Dict[str, Any], List[Any]],
    cfg: DataFetcher,
) -> ResultE[AnyType]:
    """Find a data at path or produce a value.

    return value can be:
    - value found at path
    - value found but sliced
    - value found applied to regular expression
    - conditional value depending on if statements
    - default value if all the above still produces `None`

    Flow description:

    find data from path or None ->
    apply regular expression ->
    apply slicing ->
    apply if statements ->
    return default value if Failure else mapped value
    """
    return flow(
        collection,
        partial(fetch_data_by_keys, path=cfg.path),
        lash(lambda _: Success(None)),  # type: ignore
        bind(
            partial(
                apply_regex, regex=cfg.regex,
            ),
        ),
        lash(lambda _: Success(None)),  # type: ignore
        map_(partial(
            apply_slicing, slicing=cfg.slicing,
        )),
        bind(partial(
            apply_if_statements, statements=cfg.if_statements,
        )),
        lash(  # type: ignore
            lambda _: apply_default(cfg.default),
        ),
    )


def handle_attribute(
    collection: Union[Dict[str, Any], List[Any]],
    cfg: Attribute,
) -> ResultE[AnyType]:
    """Handle one attribute with data fetchers, ifs, casting and default value.

    flow description:

    Fetch once for all data in Attribute.data_fetchers ->
    Apply separator to values if there are more than 1
    Failure -> fix to Success(None)
    Apply if statements
    Success -> Cast Value
    Failure -> apply default value

    Return Result
    """
    fetched_values = [
        fetched.unwrap()
        for fetched in  # noqa: WPS361
        [
            handle_data_fetcher(collection, data_fetcher)
            for data_fetcher in cfg.data_fetchers
        ]
        if is_successful(fetched)
    ]

    # partially declare if statement and casting functions
    ifs = partial(apply_if_statements, statements=cfg.if_statements)

    cast = safe(lambda the_value: the_value)
    if cfg.casting:
        cast = partial(apply_casting, casting=cfg.casting)  # type: ignore

    return flow(
        apply_separator(fetched_values, separator=cfg.separator),
        lash(lambda _: Success(None)),  # type: ignore
        bind(ifs),
        bind(cast),
        lash(  # type: ignore
            lambda _: apply_default(default=cfg.default),
        ),
    )



from typing import Any, List, Optional

from pydantic import ConfigDict

from kaiba.models.base import KaibaBaseModel, StrInt
from kaiba.models.if_statement import IfStatement
from kaiba.models.regex import Regex
from kaiba.models.slicing import Slicing


class DataFetcher(KaibaBaseModel):
    """Data fetcher lets you fetch data from the input."""

    path: List[StrInt] = []
    slicing: Optional[Slicing] = None
    regex: Optional[Regex] = None
    if_statements: List[IfStatement] = []
    default: Optional[Any] = None
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'path': ['path', 'to', 'data'],
            },
            {
                'path': ['path', 1, 2, 'my_val'],
                'defualt': 'if no data was found this value is used',
            },
        ],
    })



from typing import List

from pydantic import ConfigDict, Field

from kaiba.models.base import KaibaBaseModel, StrInt


class Iterator(KaibaBaseModel):
    """Allows for iterating lists at given path."""

    alias: str
    path: List[StrInt] = Field(..., min_length=1)
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'alias': 'an_item',
                'path': ['path', 'to', 10, 'data'],
            },
        ],
    })



from typing_extensions import Final

# Casting
INTEGER_CONTAINING_DECIMALS: Final[str] = 'integer_containing_decimals'
YMD_DATE_FORMAT: Final = r'(^(yy|yyyy)[^\w]?mm[^\w]?dd$)'
DMY_DATE_FORMAT: Final = r'(^dd[^\w]?mm[^\w]?(yy|yyyy)$)'
MDY_DATE_FORMAT: Final = r'(^mm[^\w]?dd[^\w]?(yy|yyyy)$)'

# Casting helpers
COMMA: Final[str] = ','
PERIOD: Final[str] = '.'
EMPTY: Final[str] = ''

# ISO
NAME: Final[str] = 'name'
ALPHA_TWO: Final[str] = 'alpha_2'
ALPHA_THREE: Final[str] = 'alpha_3'
NUMERIC: Final[str] = 'numeric'
OFFICIAL_NAME: Final[str] = 'official_name'
INVALID: Final[str] = 'invalid'



from typing import List

from pydantic import ConfigDict
from pydantic.types import StrictBool

from kaiba.models.attribute import Attribute
from kaiba.models.base import KaibaBaseModel
from kaiba.models.branching_object import BranchingObject
from kaiba.models.iterator import Iterator


class KaibaObject(KaibaBaseModel):
    """Our main object."""

    name: str
    array: StrictBool = False
    iterators: List[Iterator] = []
    attributes: List[Attribute] = []
    objects: List['KaibaObject'] = []  # noqa: WPS110
    branching_objects: List[BranchingObject] = []
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'name': 'object_name',
                'attributes': [
                    {
                        'name': 'an_attribute',
                        'default': 'a value',
                    },
                ],
            },
        ],
    })


KaibaObject.update_forward_refs()



from typing import List, Optional

from pydantic import ConfigDict

from kaiba.models.base import AnyType, KaibaBaseModel
from kaiba.models.casting import Casting
from kaiba.models.data_fetcher import DataFetcher
from kaiba.models.if_statement import IfStatement


class Attribute(KaibaBaseModel):
    """Adds an attribute with the given name."""

    name: str
    data_fetchers: List[DataFetcher] = []
    separator: str = ''
    if_statements: List[IfStatement] = []
    casting: Optional[Casting] = None
    default: Optional[AnyType] = None
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'name': 'my_attribute',
                'data_fetchers': [{
                    'path': ['abc', 0],
                }],
                'default': 'default_value',
            },
        ],
    })



from enum import Enum
from typing import Optional

from pydantic import ConfigDict

from kaiba.models.base import KaibaBaseModel


class CastToOptions(str, Enum):  # noqa: WPS600
    """Types that we can cast a value to."""

    STRING = 'string'  # noqa: WPS115
    INTEGER = 'integer'  # noqa: WPS115
    DECIMAL = 'decimal'  # noqa: WPS115
    DATE = 'date'  # noqa: WPS115


class Casting(KaibaBaseModel):
    """Allows user to cast to type."""

    to: CastToOptions
    original_format: Optional[str] = None
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                'to': 'integer',
            },
            {
                'to': 'date',
                'original_format': 'ddmmyy',
            },
            {
                'to': 'date',
                'original_format': 'yyyy.mm.dd',
            },
        ],
    })
